package avlindex

import (
	"log"

	"github.com/davecgh/go-spew/spew"

	"github.com/emirpasic/gods/utils"
)

type Node struct {
	children [2]*Node
	parent   *Node
	size     int
	value    interface{}
}

func (n *Node) String() string {
	if n == nil {
		return "nil"
	}

	p := "nil"
	if n.parent != nil {
		p = spew.Sprint(n.parent.value)
	}
	return spew.Sprint(n.value) + "(" + p + "|" + spew.Sprint(n.size) + ")"
}

type Tree struct {
	root       *Node
	comparator utils.Comparator
}

func New(comparator utils.Comparator) *Tree {
	return &Tree{comparator: comparator}
}

func (avl *Tree) String() string {
	str := "AVLTree\n"
	if avl.root == nil {
		return str + "nil"
	}
	output(avl.root, "", true, &str)
	return str
}

func (avl *Tree) Iterator() *Iterator {
	return initIterator(avl)
}

func (avl *Tree) Size() int {
	if avl.root == nil {
		return 0
	}
	return avl.root.size
}

func (avl *Tree) indexNode(idx int) *Node {
	cur := avl.root
	if idx >= 0 {
		for cur != nil {
			ls := getSize(cur.children[0])
			if idx == ls {
				return cur
			} else if idx < ls {
				cur = cur.children[0]
			} else {
				idx = idx - ls - 1
				cur = cur.children[1]
			}
		}
	} else {
		idx = -idx - 1
		for cur != nil {
			rs := getSize(cur.children[1])
			if idx == rs {
				return cur
			} else if idx < rs {
				cur = cur.children[1]
			} else {
				idx = idx - rs - 1
				cur = cur.children[0]
			}
		}
	}
	return nil
}

func (avl *Tree) Index(idx int) (interface{}, bool) {
	n := avl.indexNode(idx)
	if n != nil {
		return n.value, true
	}
	return nil, false
}

func (avl *Tree) RemoveIndex(idx int) bool {
	n := avl.indexNode(idx)
	if n != nil {
		avl.removeNode(n)
		return true
	}
	return false
}

func (avl *Tree) removeNode(n *Node) {
	if avl.root.size == 1 {
		avl.root = nil
		// return n
		return
	}

	ls, rs := getChildrenSize(n)
	if ls == 0 && rs == 0 {
		p := n.parent
		p.children[getRelationship(n)] = nil
		avl.fixRemoveHeight(p)
		// return n
		return
	}

	var cur *Node
	if ls > rs {
		cur = n.children[0]
		for cur.children[1] != nil {
			cur = cur.children[1]
		}

		cleft := cur.children[0]
		cur.parent.children[getRelationship(cur)] = cleft
		if cleft != nil {
			cleft.parent = cur.parent
		}

	} else {
		cur = n.children[1]
		for cur.children[0] != nil {
			cur = cur.children[0]
		}

		cright := cur.children[1]
		cur.parent.children[getRelationship(cur)] = cright

		if cright != nil {
			cright.parent = cur.parent
		}
	}

	cparent := cur.parent
	// 修改为interface 交换
	n.value, cur.value = cur.value, n.value

	// 考虑到刚好替换的节点是 被替换节点的孩子节点的时候, 从自身修复高度
	if cparent == n {
		avl.fixRemoveHeight(n)
	} else {
		avl.fixRemoveHeight(cparent)
	}

	// return cur
	return
}

func (avl *Tree) Remove(key interface{}) bool {

	if n, ok := avl.GetNode(key); ok {
		avl.removeNode(n)
		return true
	}
	// return nil
	return false
}

// Values 返回先序遍历的值
func (avl *Tree) Values() []interface{} {
	mszie := 0
	if avl.root != nil {
		mszie = avl.root.size
	}
	result := make([]interface{}, 0, mszie)
	avl.Traversal(func(v interface{}) bool {
		result = append(result, v)
		return true
	}, LDR)
	return result
}

func (avl *Tree) GetRange(idx1, idx2 int) (result []interface{}, ok bool) { // 0 -1

	if idx1^idx2 < 0 {
		if idx1 < 0 {
			idx1 = avl.root.size + idx1 - 1
		} else {
			idx2 = avl.root.size + idx2 - 1
		}
	}

	if idx1 > idx2 {
		ok = true
		if idx1 >= avl.root.size {
			idx1 = avl.root.size - 1
			ok = false
		}

		n := avl.indexNode(idx1)
		iter := NewIterator(n)
		result = make([]interface{}, 0, idx1-idx2)
		for i := idx2; i <= idx1; i++ {
			if iter.Next() {
				result = append(result, iter.Value())
			} else {
				ok = false
				return
			}
		}

		return

	} else {
		ok = true
		if idx2 >= avl.root.size {
			idx2 = avl.root.size - 1
			ok = false
		}

		if n := avl.indexNode(idx1); n != nil {
			iter := NewIterator(n)
			result = make([]interface{}, 0, idx2-idx1)
			for i := idx1; i <= idx2; i++ {
				if iter.Prev() {
					result = append(result, iter.Value())
				} else {
					ok = false
					return
				}
			}

			return
		}

	}

	return nil, false
}

func (avl *Tree) Get(key interface{}) (interface{}, bool) {
	n, ok := avl.GetNode(key)
	if ok {
		return n.value, true
	}
	return n, false
}

func (avl *Tree) GetAround(key interface{}) (result [3]interface{}) {
	an := avl.GetAroundNode(key)
	for i, n := range an {
		if n != nil {
			result[i] = n.value
		}
	}
	return
}

func (avl *Tree) GetAroundNode(value interface{}) (result [3]*Node) {

	if cur, ok := avl.GetNode(value); ok {

		var iter *Iterator

		iter = NewIterator(cur)
		iter.curPushPrevStack(iter.up)
		iter.up = iter.getPrevUp(iter.up)

		if v, ok := iter.tstack.Pop(); ok {
			result[0] = v.(*Node)
			// iter.curPushPrevStack(iter.cur)
		} else {
			result[0] = iter.up
		}

		iter = NewIterator(cur)
		iter.curPushNextStack(iter.up)
		iter.up = iter.getNextUp(iter.up)

		if v, ok := iter.tstack.Pop(); ok {
			result[2] = v.(*Node)
			// iter.curPushNextStack(iter.cur)
		} else {
			result[2] = iter.up
		}

		result[1] = cur
	}
	return
}
func (avl *Tree) GetNode(value interface{}) (*Node, bool) {

	for n := avl.root; n != nil; {
		switch c := avl.comparator(value, n.value); c {
		case -1:
			n = n.children[0]
		case 1:
			n = n.children[1]
		case 0:
			return n, true
		default:
			panic("Get comparator only is allowed in -1, 0, 1")
		}
	}
	return nil, false
}

func (avl *Tree) Put(value interface{}) {

	node := &Node{value: value, size: 1}
	if avl.root == nil {
		avl.root = node
		return
	}

	cur := avl.root
	parent := cur.parent
	child := -1

	for {

		if cur == nil {
			parent.children[child] = node
			node.parent = parent

			fixed := parent.parent
			fsize := getSize(fixed)
			if fsize == 3 {
				lefts, rigths := getChildrenSize(fixed)
				avl.fix3PutHeight(fixed, lefts, rigths)
			}
			return
		}

		if cur.size > 9 {
			ls, rs := cur.children[0].size, cur.children[1].size
			factor := cur.size / 10 // or factor = 1
			if rs >= ls*2+factor || ls >= rs*2+factor {
				avl.fixPutHeight(cur, ls, rs)
			}
		}

		cur.size++
		parent = cur
		c := avl.comparator(value, cur.value)
		child = (c + 2) / 2
		cur = cur.children[child]
	}
}

type TraversalMethod int

const (
	// L = left R = right D = Value(dest)
	_ TraversalMethod = iota
	//DLR 先值 然后左递归 右递归 下面同理
	DLR
	//LDR 先从左边有序访问到右边 从小到大
	LDR
	// LRD 同理
	LRD

	// DRL 同理
	DRL

	// RDL 先从右边有序访问到左边 从大到小
	RDL

	// RLD 同理
	RLD
)

// Traversal 遍历的方法
func (avl *Tree) Traversal(every func(v interface{}) bool, traversalMethod ...interface{}) {
	if avl.root == nil {
		return
	}

	method := LDR
	if len(traversalMethod) != 0 {
		method = traversalMethod[0].(TraversalMethod)
	}

	switch method {
	case DLR:
		var traverasl func(cur *Node) bool
		traverasl = func(cur *Node) bool {
			if cur == nil {
				return true
			}
			if !every(cur.value) {
				return false
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			return true
		}
		traverasl(avl.root)
	case LDR:
		var traverasl func(cur *Node) bool
		traverasl = func(cur *Node) bool {
			if cur == nil {
				return true
			}
			if !traverasl(cur.children[0]) {
				log.Println(cur)
				return false
			}
			if !every(cur.value) {
				return false
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			return true
		}
		traverasl(avl.root)
	case LRD:
		var traverasl func(cur *Node) bool
		traverasl = func(cur *Node) bool {
			if cur == nil {
				return true
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			if !every(cur.value) {
				return false
			}
			return true
		}
		traverasl(avl.root)
	case DRL:
		var traverasl func(cur *Node) bool
		traverasl = func(cur *Node) bool {
			if cur == nil {
				return true
			}
			if !every(cur.value) {
				return false
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			return true
		}
		traverasl(avl.root)
	case RDL:
		var traverasl func(cur *Node) bool
		traverasl = func(cur *Node) bool {
			if cur == nil {
				return true
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			if !every(cur.value) {
				return false
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			return true
		}
		traverasl(avl.root)
	case RLD:
		var traverasl func(cur *Node) bool
		traverasl = func(cur *Node) bool {
			if cur == nil {
				return true
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !every(cur.value) {
				return false
			}
			return true
		}
		traverasl(avl.root)
	}
}

func (avl *Tree) lrrotate3(cur *Node) {
	const l = 1
	const r = 0

	movparent := cur.children[l]
	mov := movparent.children[r]

	mov.value, cur.value = cur.value, mov.value //交换值达到, 相对位移

	cur.children[r] = mov
	mov.parent = cur

	cur.children[l] = movparent
	movparent.children[r] = nil

	cur.children[r] = mov
	mov.parent = cur

	// cur.size = 3
	// cur.children[r].size = 1
	cur.children[l].size = 1
}

func (avl *Tree) lrrotate(cur *Node) {

	const l = 1
	const r = 0

	movparent := cur.children[l]
	mov := movparent.children[r]

	mov.value, cur.value = cur.value, mov.value //交换值达到, 相对位移

	if mov.children[l] != nil {
		movparent.children[r] = mov.children[l]
		movparent.children[r].parent = movparent
		//movparent.children[r].child = l
	} else {
		movparent.children[r] = nil
	}

	if mov.children[r] != nil {
		mov.children[l] = mov.children[r]
		//mov.children[l].child = l
	} else {
		mov.children[l] = nil
	}

	if cur.children[r] != nil {
		mov.children[r] = cur.children[r]
		mov.children[r].parent = mov
	} else {
		mov.children[r] = nil
	}

	cur.children[r] = mov
	mov.parent = cur

	// cur.size = 3
	// cur.children[0].size = 1
	// cur.children[1].size = 1

	movparent.size = getChildrenSumSize(movparent) + 1
	mov.size = getChildrenSumSize(mov) + 1
	cur.size = getChildrenSumSize(cur) + 1

	// mov.height = getMaxChildrenHeight(mov) + 1
	// movparent.height = getMaxChildrenHeight(movparent) + 1
	// cur.height = getMaxChildrenHeight(cur) + 1
}

func (avl *Tree) rlrotate3(cur *Node) {
	const l = 0
	const r = 1

	movparent := cur.children[l]
	mov := movparent.children[r]

	mov.value, cur.value = cur.value, mov.value //交换值达到, 相对位移

	cur.children[r] = mov
	mov.parent = cur

	cur.children[l] = movparent
	movparent.children[r] = nil

	cur.children[r] = mov
	mov.parent = cur

	// cur.size = 3
	// cur.children[r].size = 1
	cur.children[l].size = 1
}

func (avl *Tree) rlrotate(cur *Node) {

	const l = 0
	const r = 1

	movparent := cur.children[l]
	mov := movparent.children[r]

	mov.value, cur.value = cur.value, mov.value //交换值达到, 相对位移

	if mov.children[l] != nil {
		movparent.children[r] = mov.children[l]
		movparent.children[r].parent = movparent
	} else {
		movparent.children[r] = nil
	}

	if mov.children[r] != nil {
		mov.children[l] = mov.children[r]
	} else {
		mov.children[l] = nil
	}

	if cur.children[r] != nil {
		mov.children[r] = cur.children[r]
		mov.children[r].parent = mov
	} else {
		mov.children[r] = nil
	}

	cur.children[r] = mov
	mov.parent = cur

	movparent.size = getChildrenSumSize(movparent) + 1
	mov.size = getChildrenSumSize(mov) + 1
	cur.size = getChildrenSumSize(cur) + 1
}

func (avl *Tree) rrotate3(cur *Node) {
	const l = 0
	const r = 1
	// 1 right 0 left
	mov := cur.children[l]

	mov.value, cur.value = cur.value, mov.value //交换值达到, 相对位移

	cur.children[r] = mov
	mov.size = 1

	cur.children[l] = mov.children[l]
	cur.children[l].parent = cur

	mov.children[l] = nil

	mov.size = 1
}

func (avl *Tree) rrotate(cur *Node) {

	const l = 0
	const r = 1
	// 1 right 0 left
	mov := cur.children[l]

	// lsize, rsize := getChildrenHeight(cur)
	// movrsize := getSize(mov.children[r])

	mov.value, cur.value = cur.value, mov.value //交换值达到, 相对位移

	//  mov.children[l]不可能为nil

	mov.children[l].parent = cur
	cur.children[l] = mov.children[l]

	// 解决mov节点孩子转移的问题
	if mov.children[r] != nil {
		mov.children[l] = mov.children[r]
	} else {
		mov.children[l] = nil
	}

	if cur.children[r] != nil {
		mov.children[r] = cur.children[r]
		mov.children[r].parent = mov
	} else {
		mov.children[r] = nil
	}

	// 连接转移后的节点 由于mov只是与cur交换值,parent不变
	cur.children[r] = mov

	// cur.size = 3
	// cur.children[0].size = 1
	// cur.children[1].size = 1

	mov.size = getChildrenSumSize(mov) + 1
	cur.size = getChildrenSumSize(cur) + 1
	// cur.height = getMaxChildrenHeight(cur) + 1
}

func (avl *Tree) lrotate3(cur *Node) {
	const l = 1
	const r = 0
	// 1 right 0 left
	mov := cur.children[l]

	mov.value, cur.value = cur.value, mov.value //交换值达到, 相对位移

	cur.children[r] = mov
	mov.size = 1

	cur.children[l] = mov.children[l]
	cur.children[l].parent = cur

	mov.children[l] = nil

	mov.size = 1
}

func (avl *Tree) lrotate(cur *Node) {

	const l = 1
	const r = 0
	// 1 right 0 left
	mov := cur.children[l]

	// lsize, rsize := getChildrenHeight(cur)
	// movrsize := getSize(mov.children[r])

	mov.value, cur.value = cur.value, mov.value //交换值达到, 相对位移

	//  mov.children[l]不可能为nil

	mov.children[l].parent = cur
	cur.children[l] = mov.children[l]

	// 解决mov节点孩子转移的问题
	if mov.children[r] != nil {
		mov.children[l] = mov.children[r]
	} else {
		mov.children[l] = nil
	}

	if cur.children[r] != nil {
		mov.children[r] = cur.children[r]
		mov.children[r].parent = mov
	} else {
		mov.children[r] = nil
	}

	// 连接转移后的节点 由于mov只是与cur交换值,parent不变
	cur.children[r] = mov

	// cur.size = 3
	// cur.children[0].size = 1
	// cur.children[1].size = 1

	mov.size = getChildrenSumSize(mov) + 1
	cur.size = getChildrenSumSize(cur) + 1
}

func getChildrenSumSize(cur *Node) int {
	return getSize(cur.children[0]) + getSize(cur.children[1])
}

func getChildrenSize(cur *Node) (int, int) {
	return getSize(cur.children[0]), getSize(cur.children[1])
}

func getSize(cur *Node) int {
	if cur == nil {
		return 0
	}
	return cur.size
}

func (avl *Tree) fixRemoveHeight(cur *Node) {
	for cur != nil {
		cur.size--
		if cur.size > 9 {
			ls, rs := cur.children[0].size, cur.children[1].size
			if rs >= ls*2 || ls >= rs*2 {
				avl.fixPutHeight(cur, ls, rs)
			}
		}
		cur = cur.parent
	}
}

func (avl *Tree) fix3PutHeight(cur *Node, lefts, rigths int) {
	if lefts > rigths {
		l := cur.children[0]
		llsize, lrsize := getChildrenSize(l)
		if lrsize > llsize {
			avl.rlrotate3(cur)
		} else {
			avl.rrotate3(cur)
		}
	} else {
		r := cur.children[1]
		rlsize, rrsize := getChildrenSize(r)
		if rlsize > rrsize {
			avl.lrrotate3(cur)
		} else {
			avl.lrotate3(cur)
		}
	}
}

func (avl *Tree) fixPutHeight(cur *Node, lefts, rigths int) {
	if lefts > rigths {
		l := cur.children[0]
		llsize, lrsize := getChildrenSize(l)
		if lrsize > llsize {
			avl.rlrotate(cur)
		} else {
			avl.rrotate(cur)
		}
	} else {
		r := cur.children[1]
		rlsize, rrsize := getChildrenSize(r)
		if rlsize > rrsize {
			avl.lrrotate(cur)
		} else {
			avl.lrotate(cur)
		}
	}
}

func output(node *Node, prefix string, isTail bool, str *string) {

	if node.children[1] != nil {
		newPrefix := prefix
		if isTail {
			newPrefix += "│   "
		} else {
			newPrefix += "    "
		}
		output(node.children[1], newPrefix, false, str)
	}
	*str += prefix
	if isTail {
		*str += "└── "
	} else {
		*str += "┌── "
	}

	*str += spew.Sprint(node.value) + "\n"

	if node.children[0] != nil {
		newPrefix := prefix
		if isTail {
			newPrefix += "    "
		} else {
			newPrefix += "│   "
		}
		output(node.children[0], newPrefix, true, str)
	}

}

func outputfordebug(node *Node, prefix string, isTail bool, str *string) {

	if node.children[1] != nil {
		newPrefix := prefix
		if isTail {
			newPrefix += "│   "
		} else {
			newPrefix += "    "
		}
		outputfordebug(node.children[1], newPrefix, false, str)
	}
	*str += prefix
	if isTail {
		*str += "└── "
	} else {
		*str += "┌── "
	}

	suffix := "("
	parentv := ""
	if node.parent == nil {
		parentv = "nil"
	} else {
		parentv = spew.Sprint(node.parent.value)
	}
	suffix += parentv + "|" + spew.Sprint(node.size) + ")"
	*str += spew.Sprint(node.value) + suffix + "\n"

	if node.children[0] != nil {
		newPrefix := prefix
		if isTail {
			newPrefix += "    "
		} else {
			newPrefix += "│   "
		}
		outputfordebug(node.children[0], newPrefix, true, str)
	}
}

func (avl *Tree) debugString() string {
	str := "AVLTree\n"
	if avl.root == nil {
		return str + "nil"
	}
	outputfordebug(avl.root, "", true, &str)
	return str
}
